/*
 * Copyright (C) 2012 by
 *   MetraLabs GmbH (MLAB), GERMANY
 * and
 *   Neuroinformatics and Cognitive Robotics Labs (NICR) at TU Ilmenau, GERMANY
 * All rights reserved.
 *
 * Contact: info@mira-project.org
 *
 * Commercial Usage:
 *   Licensees holding valid commercial licenses may use this file in
 *   accordance with the commercial license agreement provided with the
 *   software or, alternatively, in accordance with the terms contained in
 *   a written agreement between you and MLAB or NICR.
 *
 * GNU General Public License Usage:
 *   Alternatively, this file may be used under the terms of the GNU
 *   General Public License version 3.0 as published by the Free Software
 *   Foundation and appearing in the file LICENSE.GPL3 included in the
 *   packaging of this file. Please review the following information to
 *   ensure the GNU General Public License version 3.0 requirements will be
 *   met: http://www.gnu.org/copyleft/gpl.html.
 *   Alternatively you may (at your option) use any later version of the GNU
 *   General Public License if such license has been publicly approved by
 *   MLAB and NICR (or its successors, if any).
 *
 * IN NO EVENT SHALL "MLAB" OR "NICR" BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
 * THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF "MLAB" OR
 * "NICR" HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * "MLAB" AND "NICR" SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND "MLAB" AND "NICR" HAVE NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS.
 */

/**
 * @file RasterTransformTest.C
 *    A test case for the RasterTransformation.
 *
 * @author Christof Schröter
 * @date   2017/02/14
 */

#include <iostream>
#include <boost/test/unit_test.hpp>
#include <geometry/RasterTransformation.h>
#include <math/Random.h>
#include <image/Img.h>

using namespace std;
using namespace mira;

#define RANDOMTEST_CYCLES 1000

///////////////////////////////////////////////////////////////////////////////

BOOST_AUTO_TEST_CASE( DefinedRotationTest )
{
	Img8U3 src(1024, 768);

	// random init
	for (int y = 0; y < src.height(); ++y)
	{
		unsigned char* data = src.data(y);
		for (int x = 0; x < src.width(); ++x)
			data[x] = MIRA_RANDOM.uniform(0,256);
	}

	// tgt size is rotated
	Img8U3 tgt(src.height(), src.width());

	// just rotate by 90° (src -> tgt)
	RasterTransformation rt_fw(Rect2i(0,0,src.width()+1,src.height()+1), // using size+1 here to also get the last line
	                           Rect2i(0,0,tgt.width()+1,tgt.height()+1), // -> copy the entire image (not guaranteed otherwise)
	                           Point2d(src.width()/2,src.height()/2.f),
	                           Point2d(tgt.width()/2.f,tgt.height()/2.f),
	                           1.0,
	                           boost::math::constants::half_pi<double>(),
	                           true);

	{
		int minSrcX = 10000, maxSrcX = -1, minSrcY = 10000, maxSrcY = -1;
		int minTgtX = 10000, maxTgtX = -1, minTgtY = 10000, maxTgtY = -1;
		for (RasterTransformation::iterator it = rt_fw.begin(); it.isValid(); ++it)
		{
			minSrcX = std::min(minSrcX, it.srcX());
			maxSrcX = std::max(maxSrcX, it.srcX());
			minSrcY = std::min(minSrcY, it.srcY());
			maxSrcY = std::max(maxSrcY, it.srcY());
			minTgtX = std::min(minTgtX, it.tgtX());
			maxTgtX = std::max(maxTgtX, it.tgtX());
			minTgtY = std::min(minTgtY, it.tgtY());
			maxTgtY = std::max(maxTgtY, it.tgtY());

			tgt(it.tgtX(), it.tgtY()) = src(it.srcX(), it.srcY());
		}

		BOOST_CHECK(minSrcX == 0);
		BOOST_CHECK(maxSrcX == src.width()-1);
		BOOST_CHECK(minSrcY == 0);
		BOOST_CHECK(maxSrcY == src.height()-1);

		BOOST_CHECK(minTgtX == 0);
		BOOST_CHECK(maxTgtX == tgt.width()-1);
		BOOST_CHECK(minTgtY == 0);
		BOOST_CHECK(maxTgtY == tgt.height()-1);
	}

	// same size as original image
	Img8U3 tgt2(src.size());

	// rotate back (tgt -> tgt2)
	RasterTransformation rt_bw(Rect2i(0,0,tgt.width()+1,tgt.height()+1),
	                           Rect2i(0,0,tgt2.width()+1,tgt2.height()+1),
	                           Point2d(tgt.width()/2,tgt.height()/2),
	                           Point2d(tgt2.width()/2,tgt2.height()/2),
	                           1.0,
	                           -boost::math::constants::half_pi<double>(),
	                           true);

	{
		int minSrcX = 10000, maxSrcX = -1, minSrcY = 10000, maxSrcY = -1;
		int minTgtX = 10000, maxTgtX = -1, minTgtY = 10000, maxTgtY = -1;
		for (RasterTransformation::iterator it = rt_bw.begin(); it.isValid(); ++it)
		{
			minSrcX = std::min(minSrcX, it.srcX());
			maxSrcX = std::max(maxSrcX, it.srcX());
			minSrcY = std::min(minSrcY, it.srcY());
			maxSrcY = std::max(maxSrcY, it.srcY());
			minTgtX = std::min(minTgtX, it.tgtX());
			maxTgtX = std::max(maxTgtX, it.tgtX());
			minTgtY = std::min(minTgtY, it.tgtY());
			maxTgtY = std::max(maxTgtY, it.tgtY());

			tgt2(it.tgtX(), it.tgtY()) = tgt(it.srcX(), it.srcY());
		}
		BOOST_CHECK(minSrcX == 0);
		BOOST_CHECK(maxSrcX == tgt.width()-1);
		BOOST_CHECK(minSrcY == 0);
		BOOST_CHECK(maxSrcY == tgt.height()-1);

		BOOST_CHECK(minTgtX == 0);
		BOOST_CHECK(maxTgtX == tgt2.width()-1);
		BOOST_CHECK(minTgtY == 0);
		BOOST_CHECK(maxTgtY == tgt2.height()-1);
	}

	BOOST_CHECK(src == tgt2);

}

void info(const Rect2i& src, const Rect2i& tgt, const Point2d& srcRef, Point2d& tgtRef,
          double scale, double rot)
{
	std::cout << "src: " << src.minCorner << "-" << src.maxCorner  << std::endl;
	std::cout << "tgt: " << tgt.minCorner << "-" << tgt.maxCorner << std::endl;
	std::cout << "srcRef: " << srcRef << std::endl;
	std::cout << "tgtRef: " << tgtRef << std::endl;
	std::cout << "scale: " << scale << std::endl;
	std::cout << "rot: " << rot << " (" << rad2deg(rot) << "deg)"  << std::endl;
}

BOOST_AUTO_TEST_CASE( LimitsTest )
{
	for (int n = 1; n < RANDOMTEST_CYCLES; ++n) {
		int x, y;

		x = MIRA_RANDOM.uniform(0,100);
		y = MIRA_RANDOM.uniform(0,100);
		Rect2i src(x, y,
		           x + MIRA_RANDOM.uniform(0,100),
		           y + MIRA_RANDOM.uniform(0,100));

		x = MIRA_RANDOM.uniform(0,100);
		y = MIRA_RANDOM.uniform(0,100);
		Rect2i tgt(x, y,
		           x + MIRA_RANDOM.uniform(0,100),
		           y + MIRA_RANDOM.uniform(0,100));

		x = MIRA_RANDOM.uniform(src.x0() - src.width() * 0.1,
			                    src.x0() + src.width() * 1.1);
		y = MIRA_RANDOM.uniform(src.y0() - src.height() * 0.1,
			                    src.y0() + src.height() * 1.1);
		Point2d srcRef(x, y); // up to 10% of size outside src

		x = MIRA_RANDOM.uniform(tgt.x0() - tgt.width() * 0.1,
			                    tgt.x0() + tgt.width() * 1.1);
		y = MIRA_RANDOM.uniform(tgt.y0() - tgt.height() * 0.1,
			                    tgt.y0() + tgt.height() * 1.1);
		Point2d tgtRef(x, y); // up to 10% of size outside tgt

		double scale = MIRA_RANDOM.uniform(0.5, 2.0);
		double rot = MIRA_RANDOM.uniform(-boost::math::constants::pi<double>(),
			                             boost::math::constants::pi<double>());

		RasterTransformation rt(src, tgt, srcRef, tgtRef, scale, rot, true);

		for (RasterTransformation::iterator it = rt.begin(); it.isValid(); ++it)
		{
			if (it.srcX() < src.x0())
				info(src, tgt, srcRef, tgtRef, scale, rot);
			if (it.srcX() >= src.x1())
				info(src, tgt, srcRef, tgtRef, scale, rot);
			if (it.srcY() < src.y0())
				info(src, tgt, srcRef, tgtRef, scale, rot);
			if (it.srcY() >= src.y1())
				info(src, tgt, srcRef, tgtRef, scale, rot);

			if (it.tgtX() < tgt.x0())
				info(src, tgt, srcRef, tgtRef, scale, rot);
			if (it.tgtX() >= tgt.x1())
				info(src, tgt, srcRef, tgtRef, scale, rot);
			if (it.tgtY() < tgt.y0())
				info(src, tgt, srcRef, tgtRef, scale, rot);
			if (it.tgtY() >= tgt.y1())
				info(src, tgt, srcRef, tgtRef, scale, rot);

			BOOST_CHECK(it.srcX() >= src.x0());
			BOOST_CHECK(it.srcX() < src.x1());
			BOOST_CHECK(it.srcY() >= src.y0());
			BOOST_CHECK(it.srcY() < src.y1());

			BOOST_CHECK(it.tgtX() >= tgt.x0());
			BOOST_CHECK(it.tgtX() < tgt.x1());
			BOOST_CHECK(it.tgtY() >= tgt.y0());
			BOOST_CHECK(it.tgtY() < tgt.y1());
		}
	}
}
